#include "Circuit.h"
#include "EdgePointer.h"
#include "ParseError.h"

namespace vp
{
	Circuit::Circuit()
	{
		emptyNodesCounter = 0;
	}

//_____________________________________________________________________________________________________________________________________
	Circuit::Circuit(const string& filename)
	{
		emptyNodesCounter = 0;

		parseFile(filename);
	}

//_____________________________________________________________________________________________________________________________________
	bool Circuit::verifyFile(const string& fileName)
	{
		ifstream in(fileName.c_str());
		if (in.fail())
			return false;
		in.close();
		return true;
	}

//_____________________________________________________________________________________________________________________________________
	void Circuit::insertNode(const string& port, int revLevel)
	{
		if (nodeIndex.count(port)) throw ParseError("gate " + port + " already declared");
		nodeIndex[port] = nodes.size();
		int newNodeIndex = nodes.size();
		nodes.push_back(Node(port, this, revLevel, newNodeIndex));
		/*if (delays.count(nodes[nodes.size() - 1].getType())){
			nodes[nodes.size() - 1].setTRise(delays[nodes[nodes.size() - 1].getType()].first);
			nodes[nodes.size() - 1].setTFall(delays[nodes[nodes.size() - 1].getType()].second);
		}*/
	}

//_____________________________________________________________________________________________________________________________________
	void Circuit::parseFile(const string& fileName)
	{
		double time1, time2, time3, time4, tstart;
		time1 = 0.0;
		tstart = clock(); // Zeitmessung beginnt.
		
		if (!verifyFile(fileName)) throw ParseError("File not found");
		string line;
		ifstream inputFile(fileName.c_str());

		cout << "Parsing file has started." << endl;		
		
		inputFile >> line;
		while (line != "module") inputFile >> line;
		getline(inputFile, line);
		int openBracketIndex = line.find("(");
		moduleName = line.substr(1, openBracketIndex - 1);

		// Use regular expressions to create nodes and edges out ouf verilog file lines.

		string osp = "[\\t\\n\\r ]*";
		string sp = "[\\t\\n\\r ]+";
		string delayPart = "((?:#\\(" + osp + "([0-9.]+)" + osp + "," + osp + "([0-9.]+)" + osp + "\\))?)";
		regex wirePattern("(input|output|inout|reg|wire)" + osp + "(\\[\\d+\\:\\d+\\])?" + sp + "(\\S+)");
		//regex gatePattern("([A-Za-z0-9]+)" + sp + delayPart + osp + "([_0-9a-zA-z]+)?" + osp + "\\(([^]+)\\)");
		//regex assignPattern("assign (.+) = 1'b(0|1);");
		regex bufferPattern("(assign)" + osp + "(.+)" + osp + "=" + osp + "(.+)" + "/\\*" + "(\\d+)" + "\\*/"); 
		regex assignPattern2("(assign)" + osp + "(.+)" + osp + "=" + "(.+)" + "(&|\\||\\^|~)" + "(.+)");
		//regex assignPattern2("(assign)" + osp + "(.+)" + osp + "=" + "(.+)" + "(&|\\||\\^|~)" + "(.+)" + "/\\*" + "(\\d+)" + "\\*/"); 
		smatch match;
		while (getline(inputFile, line, ';')){

			int commentIndex = line.find("//");
			int nwlnIndex = line.find("\n");
			if (nwlnIndex != -1 && commentIndex != -1 && nwlnIndex > commentIndex){
				line.erase(commentIndex, nwlnIndex - commentIndex + 1);
			}

			if (regex_search(line, match, wirePattern)){
				// parsing the array operator[7:0]
				string arr = match[2].str();
				if (arr.size()){
					int colonIndex = arr.find(":");
					int firstNumber = atoi(arr.substr(1, colonIndex - 1).c_str());
					int secondNumber = atoi(arr.substr(colonIndex + 1, arr.size() - 1 - colonIndex).c_str());
					if (firstNumber < secondNumber) swap(firstNumber, secondNumber);
					for (int i = firstNumber; i >= secondNumber; i--)
						insertNewEdge(match[3].str() + "[" + to_string(i) + "]", match[1].str());
				}
				else
					insertNewEdge(match[3].str(), match[1].str());
			}
			else if (regex_search(line, match, assignPattern2)){
				if (match.str(4) == "&") {
					string portDescription = "(" + match.str(2) + ", " + match.str(3) + ", " + match.str(5) + ")";
					string nodeName = "A" + match.str(2);
					while (nodeName.find(" ") != string::npos) {	// Delete spaces from nodeName string.
						nodeName.erase(nodeName.find(" "), 1);
					}
					//insertNewNode(nodeName, "and", portDescription, "", "", std::stoi(match.str(6)));
					insertNewNode(nodeName, "and", portDescription, "", "", -1);
				} else if (match.str(4) == "|") {
					string portDescription = "(" + match.str(2) + ", " + match.str(3) + ", " + match.str(5) + ")";
					string nodeName = "O" + match.str(2);
					while (nodeName.find(" ") != string::npos) {	// Delete spaces from nodeName string.
						nodeName.erase(nodeName.find(" "), 1);
					} 
					//insertNewNode(nodeName, "or", portDescription, "", "", std::stoi(match.str(6)));
					insertNewNode(nodeName, "or", portDescription, "", "", -1);
				} else if (match.str(4) == "^") {
					string portDescription = "(" + match.str(2) + ", " + match.str(3) + ", " + match.str(5) + ")";
					string nodeName = "X" + match.str(2);
					while (nodeName.find(" ") != string::npos) {	// Delete spaces from nodeName string.
						nodeName.erase(nodeName.find(" "), 1);
					} 
					//insertNewNode(nodeName, "xor", portDescription, "", "", std::stoi(match.str(6)));
					insertNewNode(nodeName, "xor", portDescription, "", "", -1);
				} else if (match.str(4) == "~") {
					string portDescription = "(" + match.str(2) + ", " + match.str(5) + ")";
					string nodeName = "N" + match.str(2);
					while (nodeName.find(" ") != string::npos) {	// Delete spaces from nodeName string.
						nodeName.erase(nodeName.find(" "), 1);
					} 
					//insertNewNode(nodeName, "not", portDescription, "", "", std::stoi(match.str(6)));
					insertNewNode(nodeName, "not", portDescription, "", "", -1);
				}
			} 
			else if (regex_search(line, match, bufferPattern)){  // buffer pattern found.
				if (match.str(2) == "zeroWire " && match.str(3) == "1'b0 ") {
					//cout << "zeroWire match gefunden! " << endl;
					insertNode("zeroNode", 0);
					int nodeIndex = nodes.size() - 1;
					inputNodes.push_back(nodeIndex);
					nodes[nodeIndex].setNodeType("INPUT_PORT");
					nodes[nodeIndex].addOutput(edges.find("zeroWire"));
					nodes[nodeIndex].nIndex = nodeIndex;
					edges["zeroWire"].setSourceNode(nodeIndex);
					continue;
				}
				if (match.str(2) == "oneWire " && match.str(3) == "1'b1 ") {
					//cout << "oneWire match gefunden! " << endl;
					insertNode("oneNode", 0);
					int nodeIndex = nodes.size() - 1;
					inputNodes.push_back(nodeIndex);
					nodes[nodeIndex].setNodeType("INPUT_PORT");
					nodes[nodeIndex].addOutput(edges.find("oneWire"));
					nodes[nodeIndex].nIndex = nodeIndex;
					edges["oneWire"].setSourceNode(nodeIndex);
					continue;
				}
				string portDescription = "(" + match.str(2) + ", " + match.str(3) + ")";
				string nodeName = "B" + match.str(2) ;
				while (nodeName.find(" ") != string::npos) {	// Delete spaces from nodeName string.
						nodeName.erase(nodeName.find(" "), 1);
					} 
				//insertNewNode(nodeName, "buffer", portDescription, "", "", std::stoi(match.str(4)));
				insertNewNode(nodeName, "buffer", portDescription, "", "", -1);
				
			}
		}
		inputFile.close();
		
		std::cout << "Parse file is completed." << std::endl; 
		
		time1 += clock() - tstart; // Zeitmessung endet.
		time1 = time1/CLOCKS_PER_SEC;
		cout << "Parsing file needed time in sec.: " << time1 << endl;
		//buildAdjacencyMatrix();
		
		//*
		//  Create simple reversed topological order of all nodes in the circuit.
		time2 = 0.0;
		tstart = clock();
		
		revLevelizeCircuit();
		
		time2 += clock() - tstart; // Zeitmessung endet.
		time2 = time2/CLOCKS_PER_SEC;
		cout << "LevelizeCircuit needed time in sec.: " << time2 << endl;
		
		// Remove dead edges to simplify circuit object.
		int j = removeDeadEdges();
		cout << j << " dead edges removed." << endl;
		//*/
		
		time4 += clock() - tstart; // Zeitmessung endet.
		time4 = time4/CLOCKS_PER_SEC;
		cout << "Simplifying circuit needed time in sec.: " << time4 << endl;
		
		// Sort node list with quick sort.
		time3 = 0.0;
		tstart = clock();
		
		sortNodesByLevel();

		time3 += clock() - tstart; // Zeitmessung endet.
		time3 = time3/CLOCKS_PER_SEC;
		cout << "SortNodesByLevel() needed time in sec.: " << time3 << endl;
		//*/

	}

//_____________________________________________________________________________________________________________________________________
	void Circuit::insertNewEdge(const string& edgeName, const string& edgeType)
	{
		if (edges.count(edgeName)) throw ParseError("wire \"" + edgeName + "\" already declared");
		edges[edgeName] = Edge(edgeName, this); 
		int currEdgeSize = edgeIndex.size();
		edgeIndex[Edge(edgeName, this)] = currEdgeSize;
		EdgePointer pointer = edges.find(edgeName);
		if (edgeType == "input"){
			//inputEdgeIndex.push_back(edges.size() -1);  // Add edge index to input index list.
			//primaryEdgeIndex[edges.size() - 1] = Edge(edgeName, this);
			insertNode(edgeName, 0);
			int nodeIndex = nodes.size() - 1;
			inputNodes.push_back(nodeIndex);
			nodes[nodeIndex].setNodeType("INPUT_PORT");
			nodes[nodeIndex].addOutput(pointer);
			nodes[nodeIndex].nIndex = nodeIndex;
			edges[edgeName].setSourceNode(nodeIndex);
		}
		else if (edgeType == "output"){
			//outputEdgeIndex.push_back(edges.size() -1);  // Add edge index to output index list.
			//primaryEdgeIndex[edges.size() - 1] = Edge(edgeName, this);
			insertNode(edgeName, 0);
			int nodeIndex = nodes.size() - 1;
			outputNodes.push_back(nodeIndex);
			nodes[nodeIndex].setNodeType("OUTPUT_PORT");
			nodes[nodeIndex].addInput(pointer);
			nodes[nodeIndex].nIndex = nodeIndex;
			edges[edgeName].addDestinationNode(nodeIndex);
		}
	}


//_____________________________________________________________________________________________________________________________________
	void Circuit::removeEdge(Edge* oldEdge) {
		// Remove old edge which has no use anymore.
		oldEdge->disconnect();
		this->edges.erase(oldEdge->name);
		//this->edgeIndex.erase(*oldEdge); This causes memory access error. Leave it out for now.
	}

//_____________________________________________________________________________________________________________________________________
	void Circuit::removeNode(Node* oldNode) {
		// Note: The node will not really be removed from the nodes list because this would disqualify correct access of edges to the nodes (edges use index of node positions in nodes list). Set node type to DELETED and ignore it in further steps.
		oldNode->setNodeType("DELETED");
		oldNode->inputs.clear();
		oldNode->outputs.clear();
		this->nodeIndex.erase(oldNode->name);
	}

//_____________________________________________________________________________________________________________________________________
	int Circuit::removeDeadEdges() {
		// Edge is dead if it has no destinations.
		int removed = 0;
		map<string,Edge>::iterator it_prev;
		for (map<string,Edge>::iterator it = edges.begin(); it != edges.end(); ++it) {
			if (it->second.destinations.size() == 0) {
				//cout << "removing: " << it->first << endl;
				recursivelyRemoveEdge(&it->second);
				removed++;
				it = it_prev;
			} else {
				it_prev = it;
			}
		}
		return removed;
	}

//_____________________________________________________________________________________________________________________________________
	void Circuit::recursivelyRemoveEdge(Edge* deletedEdge) {
		Edge* input1 = NULL;
		Edge* input2 = NULL;
		if (deletedEdge->getSource() == -1) {  // Edge source was not specified. Just delete this edge.
			removeEdge(deletedEdge);
			return;
		}
		Node* deletedNode = &nodes.at(deletedEdge->getSource());
		if (deletedNode->inputs.size() > 0) {
			input1 = &*deletedNode->inputs.at(0);
			if (deletedNode->inputs.size() > 1) input2 = &*deletedNode->inputs.at(1); 
		}
		if (input1) {
			for (int i = 0; i < input1->nDestinations(); i++) {  // Erase deletedNode from input edge.
				if (nodes.at(input1->getDestination(i)).name == deletedNode->name) 
				input1->getDestinations()->erase(input1->getDestinations()->begin() + i);
			}
			if (input1->nDestinations() == 0) recursivelyRemoveEdge(input1);
		}
		if (input2) {
			for (int i = 0; i < input2->nDestinations(); i++) {  // Erase deletedNode from input edge.
				if (nodes.at(input2->getDestination(i)).name == deletedNode->name) 
				input2->getDestinations()->erase(input2->getDestinations()->begin() + i);
			}
			if (input2->nDestinations() == 0) recursivelyRemoveEdge(input2);
		}
		removeNode(deletedNode);
		removeEdge(deletedEdge);
	}

//_____________________________________________________________________________________________________________________________________
	void Circuit::insertNewNode(const string& nodeName, const string& nodeType, const string& portDescription, const string& trise, const string& tfall, int revLevel)
	{
		insertNode(nodeName == "" ? (string("GATE_") + to_string(emptyNodesCounter++)) : nodeName, revLevel);
		nodes[nodes.size() - 1].assignTask(nodeType, portDescription, trise, tfall);
		if (nodeType == "not") notNodes.push_back(nodes.size() - 1);
	}

//_____________________________________________________________________________________________________________________________________
	std::string Circuit::getModuleName() const
	{
		return moduleName;
	}

//_____________________________________________________________________________________________________________________________________
	Node& Circuit::node(size_t index)
	{
		if (index >= getNodesCount()) throw out_of_range(" (Circuit) Node index out of range");
		return nodes[index];
	}

//_____________________________________________________________________________________________________________________________________
	Node& Circuit::node(const string& nodeName)
	{
		if (!nodeIndex.count(nodeName)) throw invalid_argument("Node \"" + nodeName + "\" not found");
		return node(nodeIndex[nodeName]);
	}

//_____________________________________________________________________________________________________________________________________
	size_t Circuit::getNodesCount() const
	{
		return nodes.size();
	}

//_____________________________________________________________________________________________________________________________________
	size_t Circuit::getInputNodesCount() const
	{
		return inputNodes.size();
	}

//_____________________________________________________________________________________________________________________________________
	Node& Circuit::inputNode(size_t index)
	{
		if (index >= getInputNodesCount()) throw out_of_range("(Input) Node index out of range");
		return node(inputNodes[index]);
	}

//_____________________________________________________________________________________________________________________________________	
	vector<int> Circuit::getInputNodes() const
	{
		return inputNodes;
	}
	
//_____________________________________________________________________________________________________________________________________	
	vector<int> Circuit::getOutputNodes() const
	{
		return outputNodes;
	}

//_____________________________________________________________________________________________________________________________________
	size_t Circuit::getOutputNodesCount() const
	{
		return outputNodes.size();
	}

//_____________________________________________________________________________________________________________________________________
	Node& Circuit::outputNode(size_t index)
	{
		if (index >= getOutputNodesCount()) throw out_of_range("(Output) Node index out of range");
		return node(outputNodes[index]);
	}

//_____________________________________________________________________________________________________________________________________
	size_t Circuit::getGatesCount() const {
		return getNodesCount() - getInputNodesCount() - getOutputNodesCount();
	}

//_____________________________________________________________________________________________________________________________________
	size_t Circuit::getNodeIndex(const string& nodeName) const {
		return nodeIndex.at(nodeName);
	}
	
//_____________________________________________________________________________________________________________________________________
	void Circuit::revLevelizeCircuit() {	// Define reversed topological order of all nodes.
		if (this->getGatesCount() == 0) {
			std::cout << "Number of gate nodes is 0. End revLevel method. " << std::endl;	
			return;
		} 
		int maxLevel = 0;
		std::vector<vp::Node*> frontierNodes;
		for(int i = 0; i < this->getOutputNodesCount(); i++) {
			this->outputNode(i).revLevel = 0;
			frontierNodes.push_back(&this->outputNode(i));
		}
		for (int i = 0; i < frontierNodes.size(); i++) {
			//std::cout << "Frontier node i name: " << frontierNodes.at(i)->name << std::endl;
			for (int j = 0; j < frontierNodes.at(i)->getInputsCount(); j++) {
				//std::cout << "Frontier node j name: " << frontierNodes.at(i)->inputNode(j).name << std::endl;
				if (frontierNodes.at(i)->inputNode(j).revLevel > 0) {
					continue;
				}
				maxLevel = 0;
			
				for(int k = 0; k < frontierNodes.at(i)->inputNode(j).getOutputsCount(); k++) {
					if (frontierNodes.at(i)->inputNode(j).outputNode(k).revLevel == -1) {
						break;
					}
					if (maxLevel < (frontierNodes.at(i)->inputNode(j).outputNode(k).revLevel + 1)) {
						maxLevel = frontierNodes.at(i)->inputNode(j).outputNode(k).revLevel + 1;
					}
					if (k == frontierNodes.at(i)->inputNode(j).getOutputsCount() - 1) {
						//std::cout << "revLevel set for: " << frontierNodes.at(i)->inputNode(j).name << std::endl;
						frontierNodes.at(i)->inputNode(j).revLevel = maxLevel; 
						frontierNodes.push_back(&frontierNodes.at(i)->inputNode(j));
					}
				}
			}
		}
		std::cout << "revLevelizeCircuit() successfull" << std::endl;
	}
	
//_____________________________________________________________________________________________________________________________________	
	void Circuit::swapNodes(Node* a, Node* b) {
		Node tmp = *a; 
    		*a = *b; 
    		*b = tmp; 
	}

//_____________________________________________________________________________________________________________________________________
	size_t Circuit::partition(size_t low, size_t high) {
		Node pivot = this->sortedNodes.at(high);    // pivot 
    		size_t i = (low - 1);  // Index of smaller element 
  
    		for (size_t j = low; j <= high - 1; j++) { 
        		// If current element is smaller than or 
        		// equal to pivot 
       		 	if (this->sortedNodes.at(j).revLevel <= pivot.revLevel) { 
        		    i++;    // increment index of smaller element 
        		    swapNodes(&this->sortedNodes.at(i), &this->sortedNodes.at(j)); 
        		} 
    		} 
   		swapNodes(&this->sortedNodes.at(i+1), &this->sortedNodes.at(high)); 
    		return (i + 1); 
	}

//_____________________________________________________________________________________________________________________________________
	void Circuit::quickSort(size_t low, size_t high) {
		if (low < high) { 
        		/* pi is partitioning index, arr[p] is now 
        		   at right place */
        		size_t pi = partition(low, high); 
  		
        		// Separately sort elements before 
        		// partition and after partition 
        		this->quickSort(low, pi - 1); 
        		this->quickSort(pi + 1, high); 
    		 } 
	}
	
//_____________________________________________________________________________________________________________________________________	
	void Circuit::sortNodesByLevel() {
		// Sort nodes by level with quick sort. Save in seperate vector sortedNodes.
		this->sortedNodes.clear();
		//assert(sortedNodes.size() == 0);
		// Start with reverse nodes vector, so many nodes are already sorted correct.
		for (std::vector<Node>::reverse_iterator rit = this->nodes.rbegin(); rit != this->nodes.rend(); ++rit) {
			if (rit->type == "DELETED") continue;
			else this->sortedNodes.push_back(*rit);
		}
		//sortedNodes = std::vector<Node>(this->nodes);
		size_t start = 0;
		size_t end = this->sortedNodes.size() - 1;
		this->quickSort(start, end);
		std::cout << "sortNodesByLevel() successfull" <<std::endl;
	}

//_____________________________________________________________________________________________________________________________________
	// Write all valid circuit nodes to a file.
	void Circuit::writeCircuitToFile(const string& fileName) {
		ofstream outputFile(fileName);
		sortNodesByLevel();
		for (std::vector<Node>::reverse_iterator it = sortedNodes.rbegin(); it != sortedNodes.rend(); ++it ) {
			if (it->type == "DELETED" || it->type == "INPUT_PORT" || it->type == "OUTPUT_PORT") continue;
			outputFile << "assign " << it->outputs.at(0)->name << " = ";
			if (it->type == "not") {
				outputFile << "~" << it->inputs.at(0)->name << " /*" << it->revLevel << "*/" << endl;
				continue;
			} else {
				outputFile << it->inputs.at(0)->name;
				if (it->type == "and") outputFile << " & " << it->inputs.at(1)->name << " /*" << it->revLevel << "*/" << endl;
				else if (it->type == "or") outputFile << " | " << it->inputs.at(1)->name << " /*" << it->revLevel << "*/" << endl;
				else if (it->type == "xor") outputFile << " ^ " << it->inputs.at(1)->name << " /*" << it->revLevel << "*/" << endl;
			}  
		}
		outputFile.close();
	}	
}
