#pragma once
#include <string>
#include <fstream>
#include <iostream>
#include <vector>
#include <map>
#include <set>
#include <regex>
#include <queue>
#include "time.h"
#include "Node.h"
#include "Edge.h"

using namespace std;

namespace vp
{

	class Circuit
	{
	
	public:
		bool verifyFile(const string&);
		void insertNode(const string&, int);
		void insertNewEdge(const string&, const string&);
		void insertNewNode(const string&, const string&, const string&, const string&, const string&, int);
		vector<int> outputNodes;  // Indices of all output nodes.
		vector<int> inputNodes;	  // Indices of all input nodes.
		vector<Node> nodes;	 // List of all nodes in the circuit.
		vector<Node> sortedNodes;  // Seperate list of all nodes sorted.
		vector<int> notNodes;  // Extra list of all NOT nodes.
		string moduleName;
		map<string, int> nodeIndex;  // Get index (pos. of node in nodes vector) to a specific node name. Not used anymore since node objects have their own index member variable.
		map<string, Edge> edges;  // Map of edge names to edge objects.
		map<Edge, int> edgeIndex;  // Map of edges to edge indexes which are used as polynomial variables.
		int emptyNodesCounter;
		
	public:
		friend class Node;
		friend class Edge;
		friend class Verifizierer;

		Circuit();
		Circuit(const string&);
		Circuit(const string&, const string&);
		void parseFile(const string&);	// Gets the path to the Verilog gate level netlist file.
		void revLevelizeCircuit();	// Create simple reversed topological order of all nodes.
		
		void writeCircuitToFile(const string&);  // Write circuit nodes to a specified verilog file.
		
		void swapNodes(Node* a, Node* b); // Used for quick sort. 
		size_t partition (size_t low, size_t high);  // Used for quick sort. 
		void quickSort(size_t low, size_t high);  // Quick sort.
		void sortNodesByLevel();  // Sort nodes by their revLevel using quicksort algorithm.
		
		void removeEdge(Edge* oldEdge);  // Remove dead edge.
		void removeNode(Node* oldNode);	 // Remove dead node.
		int removeDeadEdges();  // Check all edges whether they can be removed or not.
		void recursivelyRemoveEdge(Edge* deletedEdge);  // Remove edge recursively.

		vector<int> getInputNodes() const;				// Returns a vector containing input nodes.
		vector<int> getOutputNodes() const;

		string getModuleName() const;						// Returns the Verilog module name

		Node& node(size_t index);
		Node& node(const string& name);
		size_t getNodesCount() const;
		size_t getNodeIndex(const string& nodeName) const;	// Returns the index of a node

		Node& outputNode(size_t index);
		size_t getOutputNodesCount() const;

		Node& inputNode(size_t index);
		size_t getInputNodesCount() const;

		size_t getGatesCount() const;
	};

}
