/*
 * Monom2.cpp
 *
 *  Created on: 30.04.2018
 *      Author: Alexander Konrad
 */

#include "Monom2.h"

// _______________________________________________________________________________________________________________________________________
Monom2::Monom2(){  // Empty constructor
	this->vars = 0;
	this->ptrs = 0;
	this->size = 0;
	this->sum = 0;
	this->factor = 0;
}

// _______________________________________________________________________________________________________________________________________
Monom2::Monom2(const Monom2& old){  // Copy constructor
	this->vars = new int[old.size];
	this->ptrs = new MyList::ListElement*[old.size];
	std::memcpy(this->vars, old.vars, old.size * sizeof(varIndex));
	std::memcpy(this->ptrs, old.ptrs, old.size * sizeof(MyList::ListElement*));
	this->size = old.size;
	this->sum = old.sum;
	this->factor = old.factor;
}

// _______________________________________________________________________________________________________________________________________
Monom2& Monom2::operator= (const Monom2 &old)  // Assignment operator.
{ 
   	// Check for self assignment 
   	if(this != &old) {
   		delete[] this->vars;
		delete[] this->ptrs;
		this->vars = new int[old.size];
		this->ptrs = new MyList::ListElement*[old.size];
		std::memcpy(this->vars, old.vars, old.size * sizeof(varIndex));
		std::memcpy(this->ptrs, old.ptrs, old.size * sizeof(MyList::ListElement*));
		this->size = old.size;
		this->sum = old.sum;
		this->factor = old.factor;     
  	}
   	return *this; 
} 

// _______________________________________________________________________________________________________________________________________
Monom2::Monom2(varIndex index) {
	// create new array representating the monom
	this->vars = new int[1];
	this->ptrs = new MyList::ListElement*[1];
	this->vars[0] = index;
	this->size = 1;
	this->sum = index;
	this->factor = 1;
}

// _______________________________________________________________________________________________________________________________________
Monom2::Monom2(varIndex myints[], int size) {
	// TODO: Order varIndices from low to high and eliminate duplicatas. Monomials should not contain duplicates and should always be ordered.
	// create new array representating the monom
	this->vars = new int[size];
	this->ptrs = new MyList::ListElement*[size];
	for (int i=0; i < size; i++) {
		this->vars[i] = myints[i];
	}
	this->size = size;
	this->sum = this->calculateSum();
	this->factor = 1;
}

// _______________________________________________________________________________________________________________________________________
Monom2::Monom2(varIndex myints[], int size, int sum, mpz_class factor) {
	// create new array representating the monom.
	this->vars = new int[size];
	this->ptrs = new MyList::ListElement*[size];
	for (int i=0; i < size; i++) {
		this->vars[i] = myints[i];
	}
	this->size = size;
	this->sum =  sum;
	this->factor = factor;
}

// _______________________________________________________________________________________________________________________________________
Monom2::~Monom2() {
	delete[] this->vars;
	delete[] this->ptrs;
}

// _______________________________________________________________________________________________________________________________________
varIndex* Monom2::getVars() const{
	return this->vars;
}

// _______________________________________________________________________________________________________________________________________
MyList::ListElement** Monom2::getPtrs() const{
	return this->ptrs;
}

// _______________________________________________________________________________________________________________________________________
void Monom2::setPtrs(MyList::ListElement** ptrs) {
	this->ptrs = ptrs;
}

// _______________________________________________________________________________________________________________________________________
int Monom2::getSize() const {
	return this->size;
}

// _______________________________________________________________________________________________________________________________________
void Monom2::setSize(int size) {
	this->size = size;
}

// _______________________________________________________________________________________________________________________________________
mpz_class Monom2::getFactor() const {
	return this->factor;
}

// _______________________________________________________________________________________________________________________________________
void Monom2::setFactor(mpz_class fact) {
	this->factor = fact;
}

// _______________________________________________________________________________________________________________________________________
int Monom2::getSum() const {
	return this->sum;
}

// _______________________________________________________________________________________________________________________________________
void Monom2::setSum(int newSum) {
	this->sum = newSum;
}

// _______________________________________________________________________________________________________________________________________
int Monom2::calculateSum() const{
	int result = 0;
	for (int i = 0; i < this->size; i++) {
		result += this->vars[i];
	}
	return result;
}

// _______________________________________________________________________________________________________________________________________
bool Monom2::operator>(const Monom2 &m1) const {
	return (m1 < *this);
	//return false;
}

// _______________________________________________________________________________________________________________________________________
bool Monom2::operator<(const Monom2 &m1) const {
	if (this->getSum() != m1.getSum()) return this->getSum() < m1.getSum();
	if (this->getSize() != m1.getSize()) return this->getSize() < m1.getSize(); 
	for (int i=0; i < this->getSize(); i++) {
		if (this->vars[i] != m1.vars[i]) return this->vars[i] < m1.vars[i];	
	}
	return false;
}

// _______________________________________________________________________________________________________________________________________
bool Monom2::operator==(const Monom2 &m1) const {
	if (this->getSize() != m1.getSize()) return false;
	if (this->getSum() != m1.getSum()) return false;	
	for (int i=0; i < this->getSize(); i++) {
		if (!(this->vars[i] == m1.vars[i])) return false;
	}
	return true;
}

bool Monom2::operator!=(const Monom2 &m1) const {
	return !(*this == m1);
}

// _______________________________________________________________________________________________________________________________________
std::ostream& operator<<(std::ostream& stdout, const Monom2& obj) {
	//std::string start = "[" + std::string(mpz_get_str(NULL, 0, obj.getFactor().get_mpz_t())) + "*" , end = "]", delim = "*";
	std::string start = "[" + obj.getFactor().get_str() + "*" , end = "]", delim = "*";
	std::string s;
    	if (obj.getSize() > 0){
        	s += start;
        	for (int i = 0; i < obj.getSize(); i++){
            	s += "x" + std::to_string(obj.vars[i]);
            	//s += Integer.toString(obj.vars[i]);
            	if (i != obj.getSize() - 1) s += delim;
        	}
        	s += end;
        	//Add factor at the end
        
    	}
    	else{
        s += start;
        s += end;
    	}
   	stdout << s;
   	return stdout;

}

// _______________________________________________________________________________________________________________________________________
const std::string Monom2::to_string() const {
    std::ostringstream ss;
    ss << *this;
    return ss.str();
}

// _______________________________________________________________________________________________________________________________________
Monom2 Monom2::merge(varIndex replace, Monom2 mon) {
	// TODO: Wenn man Monome über eine Variable merged, die nicht vorkommt, dann ist Ergebnis eins zu kurz.
	int newSize = this->size + mon.getSize() - 1; // Calculate new monom length.
	if (newSize < 0) { // Merge empty monoms.
		Monom2 empty;
		empty.setFactor(this->factor * mon.getFactor());
		return empty;
	}
	varIndex* result = new int[newSize]; // Create new array for saving monom temporary.
	varIndex one = -1; // Used for variable indexes.
	varIndex two = -1;
	int pos1 = 0; // Used for current position in monomial. 
	int pos2 = 0;
	int equal = 0; // Count how often an index is doubled.
	bool oneEnd = false; // Notice end of monomials.
	bool twoEnd = false;
	bool oneEmpty = false;  //Notice if one is the empty monom(i.e. just a coefficent without variables).
	bool twoEmpty = false; 
	int sum = 0;  // While merging, calculate sum simultaneously.
	// NOTE: "one" is the monomial object which called this function, "two" is monomial object given as "mon".
	if (this->size == 0) {oneEmpty = true;}
	if (mon.getSize() == 0) {twoEmpty = true;}
	for(int i = 0; i < newSize; i++) {
		if (!oneEmpty) {
			one = this->vars[pos1];
			if (one == replace) { // Do not add replaced variable.
				pos1 += 1;
				if (pos1 == this->size) { 
			 		oneEnd = true;
			 		pos1--;
				} else {	 
					one = this->vars[pos1];
				}
			}
		} else {
			oneEnd = true;
		}
		if (!twoEmpty) {
			two = mon.vars[pos2];
			if (two == replace) { // Do not add replaced variable.
				pos2 += 1;
				if (pos2 == mon.getSize()) { 
					twoEnd = true;
				 	pos2--;
				} else {	
					two = mon.vars[pos2];
				}
			}
		} else {
			twoEnd = true;
		}

		if (!oneEnd && !twoEnd) {
			if (one == two) { // Both monomials have current variable. Only need to copy one variable into new Monom.
				result[i - equal] = one;  // Copy variable index.
				sum += one;  // Calculate sum simultaneously.
				equal++;
			 	i++;
			 	pos1++;
			 	pos2++;
			 	if (pos1 == this->size) { 
			 		oneEnd = true;
			 		pos1--;
			 	}	 
			 	if (pos2 == mon.getSize()) { 
			 		twoEnd = true;
			 		pos2--;	
			 	}	
			} else { 
				if (one < two) {
		 			result[i - equal] = one;
		 			sum += one;  // Calculate sum simultaneously.
		 			pos1++;
		 			if (pos1 == this->size) { 
		 				//std::cout << "POS1 MAX ERREICHT" << '\n';
			 			oneEnd = true;
			 			pos1--;
			 		}	
				} else {
					result[i - equal] = two; 
					sum += two;  // Calculate sum simultaneously.
					pos2++;
					if (pos2 == mon.getSize()) { 
			 			twoEnd = true;
			 			pos2--;	
			 		}
				}
			}
		} else { // One of monoms already copied completely.
			if (oneEnd) {
				result[i - equal] = two;
				sum += two;  // Calculate sum simultaneously.
				if (one == two) {
					equal++;
				 	i++;
				 	pos2++;
				} else { 
					pos2++;
				}
			} else if (twoEnd) {
				result[i - equal] = one;
				sum += one;  // Calculate sum simultaneously.
				pos1++;
				if (one == two) {	
					equal++;
				 	i++;
				}
			}
		}	
		//std::cout << "result: " << result[i] << '\n';
		
	}
	// Create new merged monomial object and return it.
	Monom2 nMon(result, newSize - equal, sum, this->factor * mon.getFactor());
	
	// Speicher wieder freigeben.
	delete[] result;
	
	return nMon;
}

// _______________________________________________________________________________________________________________________________________
Monom2 Monom2::multiply(Monom2 mon1, Monom2 mon2) {
	int newSize = mon1.getSize() + mon2.getSize(); // Calculate new monom length.
	if (newSize < 0) { // Merge empty monoms.
		Monom2 empty;
		empty.setFactor(mon1.getFactor() * mon2.getFactor());
		return empty;
	}
	varIndex* result = new int[newSize]; // Create new array for saving monom temporary.
	varIndex one = -1; // Used for variable indexes.
	varIndex two = -1;
	int pos1 = 0; // Used for current position in monom. 
	int pos2 = 0;
	int equal = 0; // Count how often an index is doubled.
	bool oneEnd = false; // Notice end of monoms.
	bool twoEnd = false;
	bool oneEmpty = false;  //Notice if one is the empty monom (e.g. just a factor).
	bool twoEmpty = false; 
	int sum = 0;  // While merging, calculate sum simultaneously.
	if (mon1.getSize() == 0) {oneEmpty = true;}
	if (mon2.getSize() == 0) {twoEmpty = true;}
	for(int i = 0; i < newSize; i++) {
		//std::cout << "Schritt: " << i << '\n'; 
		if (!oneEmpty) {
			one = mon1.vars[pos1];
		} else {
			oneEnd = true;
		}
		
		if (!twoEmpty) {
			two = mon2.vars[pos2];
		} else {
			twoEnd = true;
		}

		if (!oneEnd && !twoEnd) {
			if (one == two) { // Both have current variable. Only need to copy one variable into new Monom.
				result[i - equal] = one;  // Copy variable index.
				sum += one;  // Calculate sum simultaneously.
				equal++;
			 	i++;
			 	pos1++;
			 	pos2++;
			 	if (pos1 == mon1.size) { 
			 		oneEnd = true;
			 		pos1--;
			 	}	 
			 	if (pos2 == mon2.getSize()) { 
			 		twoEnd = true;
			 		pos2--;	
			 	}	
			} else { 
				if (one < two) {
		 			result[i - equal] = one;
		 			sum += one;  // Calculate sum simultaneously.
		 			pos1++;
		 			if (pos1 == mon1.size) { 
			 			oneEnd = true;
			 			pos1--;
			 		}	
				} else {
					result[i - equal] = two; 
					sum += two;  // Calculate sum simultaneously.
					pos2++;
					if (pos2 == mon2.getSize()) { 
			 			twoEnd = true;
			 			pos2--;	
			 		}
				}
			}
		} else { // One of monoms already copied completely.
			if (oneEnd) {
				result[i - equal] = two;
				sum += two;  // Calculate sum simultaneously.
				if (one == two) {
					equal++;
				 	i++;
				 	pos2++;
				} else { 
					pos2++;
				}
			} else if (twoEnd) {
				result[i - equal] = one;
				sum += one;  // Calculate sum simultaneously.
				pos1++;
				if (one == two) {	
					equal++;
				 	i++;
				}
			}
		}	
		//std::cout << "result: " << result[i] << '\n';
		
	}
	//std::cout << "equal =" << equal << '\n';
	Monom2 nMon(result, newSize - equal, sum, mon1.factor * mon2.getFactor());
	//nMon.factor = this->factor * mon.getFactor();
	// Speicher wieder freigeben.
	delete[] result;
	//delete newPtrs;
	return nMon;
}
