/*
 * Monom2.h
 *
 *  Created on: 30.04.2018
 *      Author: Alexander Konrad
 */

#include <string>
#include <stdlib.h>
#include <cstring>
#include <iostream>
#include <sstream>
#include <gmpxx.h>
#include "MyList.h"

#ifndef MONOM2_H_
#define MONOM2_H_

typedef int varIndex;

class Monom2 {

friend class Polynom;
public:	
	Monom2();
	Monom2(const Monom2& old);  // Copy constructor.
	Monom2& operator = (const Monom2& old); // assignment operator
	Monom2(varIndex index);  // Create monomial with one variable, index is the variable index.
	Monom2(varIndex myints[], int size);  // Create monomial from array of indices. ATTENTION: Array must be sorted from low to high and must not contain duplicates.
	//Monom2(varIndex myints[], int size, int sum, int64_t factor);
	Monom2(varIndex myints[], int size, int sum, mpz_class factor); // Create monomial from array of indices with already given coefficient.
	virtual ~Monom2();
	varIndex* getVars() const;
	MyList::ListElement** getPtrs() const;  // Get pointers of list elements in which location of each monomial is stored to use in polynomials.
	void setPtrs(MyList::ListElement** ptrs);
	int getSize() const;
	void setSize(int size);
	//int64_t getFactor() const;
	mpz_class getFactor() const;
	//void setFactor(int64_t fact);
	void setFactor(mpz_class fact);
	int getSum() const;
	void setSum(int newSum);
	int calculateSum() const;
	const std::string to_string() const;
	Monom2 merge(varIndex replace, Monom2 mon);  // Create and return new monomial by replacing variable "replace" with a whole monomial "mon".
	static Monom2 multiply(Monom2 mon1, Monom2 mon2);
	bool operator>(const Monom2 &m1) const;
	bool operator<(const Monom2 &m1) const;
	bool operator==(const Monom2 &m1) const;
	bool operator!=(const Monom2 &m1) const;
	friend std::ostream& operator<<(std::ostream& stdout, const Monom2& obj);
	
private:
	varIndex *vars;  // Array of all variable indices contained in the monomial.
	int size;  // Size of the monomial.
	int sum;  // Sum of variable indices. Used for quick comparison between monomials.
	mutable mpz_class factor;  // Coefficient of the monomial. 
	MyList::ListElement** ptrs;  // List of pointer elements to MyList entries which contains pointers of monomials. MyList is used by polynomial class for managing list of pointers to every monomial for constant access and delete time.

};

#endif /* MONOM2_H_ */
