#pragma once
#include <string>
#include <fstream>
#include <iostream>
#include <vector>
#include <map>
#include <set>
#include <regex>
#include <queue>
#include "EdgePointer.h"
using namespace std;

namespace vp
{
	class Circuit;
	class Edge;
	/*
		Node class represents an input/output port or a gate
		*/

	class Node
	{
	public:
		Circuit* circuit;
		double tRise, tFall;
		int adder = -1;  // Not used in simple version.
		int nIndex;  // Node index.

		string type;  // Type of node. ENUMS would be better instead of strings.
		string name;  // Name of the node.
		int revLevel;  // Position in reverse topological order of the circuit.
		
		void setNodeType(string);
		void addOutput(EdgePointer);
		void addInput(EdgePointer);
		void assignTask(const string&, const string&, const string&, const string&);

		Node();
		Node(string, Circuit*);
		Node(string, Circuit*, int, int);

	public:
		friend class Circuit;
		friend class Edge;
		friend struct fullAdder;

		vector<EdgePointer> inputs;
		vector<EdgePointer> outputs;

		Node(const Node&);
		Node& operator=(const Node&);

		string getName() const;				// Returns the input or gate name
		string getType() const;				// Returns the type of the gate

		bool isInputPort() const;
		bool isOutputPort() const;
		bool isGate() const;

		Node& inputNode(size_t index);
		Node& outputNode(size_t index);

		size_t getInputsCount();
		size_t getOutputsCount();

		double getTRise() const;
		double getTFall() const;
		void setTRise(double d);
		void setTFall(double d);
	};

}
