/*
 * Polynom.cpp
 *
 *  Created on: 10.08.2018
 *      Author: Alexander Konrad
 */

#include "Polynom.h"


// ________________________________________________________________________________________________________________________________________
Polynom::Polynom(){
	this->refList = new MyList[1];
	this->varSize = 1; 
}

// ________________________________________________________________________________________________________________________________________
Polynom::Polynom(int varSize){
	this->refList = new MyList[varSize];
	this->varSize = varSize;
}

/* ________________________________________________________________________________________________________________________________________
Polynom::Polynom(Monom2 mon){
	this->polySet.insert(mon);
}
*/

// ________________________________________________________________________________________________________________________________________
Polynom::Polynom(const Polynom& old){  // Copy constructor
	this->varSize = old.varSize;
	this->refList = new MyList[old.varSize];
	this->polySet.clear(); 
	for (std::set<Monom2>::iterator it = old.polySet.begin(); it != old.polySet.end(); ++it) {
		this->addMonom(*it);
	}
}

// ________________________________________________________________________________________________________________________________________
Polynom& Polynom::operator=(const Polynom& other)  // Assignment operator.
	{
		if (this != &other) {
			this->varSize = other.varSize;
			delete[] this->refList;
			this->refList = new MyList[other.varSize];
			this->polySet.clear(); 
			for (std::set<Monom2>::iterator it = other.polySet.begin(); it != other.polySet.end(); ++it) {
				this->addMonom(*it);
			}
		}
		return *this;
	}

// ________________________________________________________________________________________________________________________________________
Polynom::~Polynom(){
	delete[] this->refList;
}

// ________________________________________________________________________________________________________________________________________
std::ostream& operator<<(std::ostream& stdout, const Polynom& obj){
    std::string start = "", end = "", delim = " + ";
    std::string s;
    int size = obj.getSet()->size();
    if (!obj.getSet()->empty()){
        s += start;
        int num = 0;
        for (std::set<Monom2>::iterator it=obj.getSet()->begin(); it != obj.getSet()->end(); ++it) {
            s += (*it).to_string();
            num += 1;
            //s += Integer.toString(obj.vars[i]);
            if (num != size) s += delim;
        }
        s += end;
        //Add factor at the end
        //std::string fact = "[" + std::to_string(obj.getFactor()) + "]";
        //s += fact;
    }
    else{
        //s += start;
        //s += end;
        s += "0 (empty Polynomial)";
    }
    stdout << s;
    return stdout;
}

// ________________________________________________________________________________________________________________________________________
bool Polynom::addPolynom(const Polynom& other) {
	if (this->varSize < other.varSize) {
		std::cout << "Cant add big polynom to small polynom (considering variable range). " << std::endl;
		return false;
	} else {
		for (std::set<Monom2>::iterator it = other.polySet.begin(); it != other.polySet.end(); ++it) {
				this->addMonom(*it);
		}
		return true;
	}
}

// ________________________________________________________________________________________________________________________________________
void Polynom::addMonom(Monom2 mon){
	std::pair<std::set<Monom2>::iterator,bool> ret;
	ret = this->polySet.insert(mon);  // Insert new monomial.
	if (ret.second == false) {  //Monomial already exists. Just add the factor. Check for monomials with factor 0.
		if ((ret.first->getFactor() + mon.getFactor()) == 0) {
			//std::cout << "Monom with factor 0 erased:" << *ret.first << '\n';
			this->eraseMonom(*ret.first);
		} else {
			ret.first->factor = ret.first->getFactor() + mon.getFactor();  // Change factor of monomial.
		}
	} else { // New monom inserted.
		varIndex* vars = ret.first->getVars();
		int size = ret.first->getSize();
		for (int i = 0; i < size; i++) {
			// Add reference for this new monomial into refList.
			this->addRefVar(const_cast<Monom2&>(*ret.first), vars[i], i);
		}	
	}
}

// ________________________________________________________________________________________________________________________________________
void Polynom::eraseMonom(Monom2 mon) {
	varIndex var = 0;
	for (int i = 0; i < mon.getSize(); i++) {  // For every variable in monomial, get its varIndex and delete its entry in refList.
		var = (mon.getVars())[i];
		this->refList[var].deleteElement(mon.getPtrs()[i]);
		
	}
	size_t deletedElements = 0;	
	deletedElements = this->polySet.erase(mon);  // Delete monomial from polynom.
	//std::cout << "Monom erased: " << mon << " Amount: " << deletedElements << '\n';
}

// ________________________________________________________________________________________________________________________________________
void Polynom::replaceAND(varIndex replace, varIndex in1, varIndex in2) {
	varIndex tmp = -1;
	if (in1 == in2) {
		Monom2 oneMonom(in1);
		std::list<Monom2> mons;
		mons.push_back(oneMonom);
		this->replace(replace, mons);
	} else {		// Normal case. Create monomials representing AND gate: replace = in1*in2.
		if (in1 > in2) {
			tmp = in1;
			in1 = in2;
			in2 = tmp;
		}	
		varIndex* vars = new int[2];
		vars[0] = in1;
		vars[1] = in2;
		Monom2 andMon(vars, 2);
		std::list<Monom2> mons;
		mons.push_back(andMon);
		this->replace(replace, mons);
		delete[] vars;
	}
}

// ________________________________________________________________________________________________________________________________________
void Polynom::replaceOR(varIndex replace, varIndex in1, varIndex in2) {
	varIndex tmp = -1;
	if (in1 == in2) {  // Special case. OR with same inputs.
		Monom2 oneMonom(in1);
		std::list<Monom2> mons;
		mons.push_back(oneMonom);
		this->replace(replace, mons);
	} else {		// Normal case. Create monomials representing OR gate: replace = in1 + in2 - in1*in2.
		if (in1 > in2) {
			tmp = in1;
			in1 = in2;
			in2 = tmp;		
		}	
		varIndex* vars = new int[2];
		vars[0] = in1;
		vars[1] = in2;
		Monom2 orMon1(in1);
		Monom2 orMon2(in2);
		Monom2 orMon3(vars, 2);
		orMon3.setFactor(-1);
		std::list<Monom2> mons;
		mons.push_back(orMon1);
		mons.push_back(orMon2);
		mons.push_back(orMon3);
		this->replace(replace, mons);
		delete[] vars;
	}
}

// ________________________________________________________________________________________________________________________________________
void Polynom::replaceXOR(varIndex replace, varIndex in1, varIndex in2) {
	varIndex tmp = -1;
	if (in1 == in2) {  // Special case. XOR with same inputs.
		Monom2 empty;
		empty.setFactor(0);
		std::list<Monom2> mons;
		mons.push_back(empty);
		this->replace(replace, mons);
	} else {		// Normal case. Create monomials representing xor gate: replace = in1 + in2 - 2*in1*in2.
		if (in1 > in2) {
			tmp = in1;
			in1 = in2;
			in2 = tmp;
		}	
		varIndex* vars = new int[2];
		vars[0] = in1;
		vars[1] = in2;
		Monom2 xorMon1(in1);
		Monom2 xorMon2(in2);
		Monom2 xorMon3(vars, 2);
		xorMon3.setFactor(-2);
		std::list<Monom2> mons;
		mons.push_back(xorMon1);
		mons.push_back(xorMon2);
		mons.push_back(xorMon3);
		this->replace(replace, mons);
		delete[] vars;
	}
}

// ________________________________________________________________________________________________________________________________________
void Polynom::replaceNOT(varIndex replace, varIndex in1) {
	Monom2 notMon1(in1);
	notMon1.setFactor(-1);
	Monom2 notMon2;  // This is constant 1 monomial.
	notMon2.setFactor(1);
	std::list<Monom2> mons;
	mons.push_back(notMon1);
	mons.push_back(notMon2);
	this->replace(replace, mons);
}

// ________________________________________________________________________________________________________________________________________
void Polynom::replaceBUFFER(varIndex replace, varIndex in1) {
	Monom2 notMon1(in1);
	notMon1.setFactor(1);
	std::list<Monom2> mons;
	mons.push_back(notMon1);
	this->replace(replace, mons);
}
/*
// ________________________________________________________________________________________________________________________________________
void Polynom::replaceCON0(varIndex replace) {
	Monom2 notMon1;
	notMon1.setFactor(0);
	std::list<Monom2> mons;
	mons.push_back(notMon1);
	this->replace(replace, mons);
}
*/
// ________________________________________________________________________________________________________________________________________
void Polynom::replace(varIndex replace, std::list<Monom2> mons) {
	Monom2 newMon;
	Monom2 oldMon;
	MyList::ListElement* nextElement;
	for (MyList::Iterator it=this->refList[replace].begin(); it != this->refList[replace].end(); it = nextElement) {
		if (refList[replace].isEmpty()) {
			std::cout << "Reflist Empty. Something went wrong." << std::endl;
			return;
		} 
		oldMon = *(it.returnData());
		nextElement = it.returnElement()->next;  // Save address of next element before deleting element.
		this->eraseMonom(*(it.returnData()));  // ATTENTION: Iterator on the current element gets invalid. 
		it = this->refList[replace].begin();
		for (std::list<Monom2>::iterator it2=mons.begin(); it2 != mons.end(); ++it2) {
			newMon = (*it2).merge(replace, oldMon);
			//std::cout << "Neues Monom hinzugefügt: " << newMon << '\n';
			if (newMon.getFactor() == 0) continue; // Dont add monom with factor 0. Only caused by XOR with same inputs.
			this->addMonom(newMon);	
		}
	
	}
}

// ________________________________________________________________________________________________________________________________________
void Polynom::addRefVar(Monom2& mon, varIndex index, int i) {
	(mon.ptrs)[i] = this->refList[index].add(&mon);
	//(mon.ptrs)[i] = (&(this->refList[index].returnTail())); 
	return;
}

// ________________________________________________________________________________________________________________________________________
const std::set<Monom2>* Polynom::getSet() const {
	return &this->polySet;
}

// ________________________________________________________________________________________________________________________________________
MyList* Polynom::getRefList() {
	return this->refList;
}

// ________________________________________________________________________________________________________________________________________
size_t Polynom::getVarSize() {
	return this->varSize;
}

// ________________________________________________________________________________________________________________________________________
size_t Polynom::size() {
	return this->polySet.size();
}

// ________________________________________________________________________________________________________________________________________
void Polynom::resize(size_t varSize) {
	delete[] this->refList;
	this->polySet.clear();
	this->refList = new MyList[varSize];
	this->varSize = varSize;
}

// _______________________________________________________________________________________________________________________________________
Polynom Polynom::multiplyPoly(Polynom& p1, Polynom& p2) {
	int maxSize = 0;
	if (p1.getVarSize() < p2.getVarSize()) {
		maxSize = p2.getVarSize();
	} else {
		maxSize = p1.getVarSize();
	}
	Polynom mult(maxSize);
	
	Monom2 temp;
	for (std::set<Monom2>::iterator it= p1.polySet.begin(); it != p1.polySet.end(); ++it) {
		for (std::set<Monom2>::iterator it2= p2.polySet.begin(); it2 != p2.polySet.end(); ++it2) {
			temp = Monom2::multiply(*it, *it2);
			mult.addMonom(temp);
		}
	}
	
	return mult;
}
