/*
 * Verifizierer.cpp
 *
 *  Created on: 08.01.2019
 *      Author: Alexander Konrad
 */


#include "Verifizierer.h"


// _______________________________________________________________________________________________________________________________________
Verifizierer::Verifizierer(const string& fileName, const string& sigFile){  // Construct from a parsed circuit file.
	// Initialize Circuit

	try {
		//Parse circuit file
		this->circuit.parseFile(fileName);

		
	}
	catch (vp::ParseError& e) {
		std::cout << "ParseError: " << e.what() << std::endl;
		exit(EXIT_FAILURE);
		//return;
	}
	catch (exception& e) {
		std::cout << e.what() << std::endl;
		exit(EXIT_FAILURE);
	}
	// Initialize polynom, read signature file and create signature polynom.
	//this->poly = Polynom(this->circuit.edgeIndex.size());
	//this->poly.addPolynom(createSignaturePolynom(readSignatureFile(sigFile)));
}

// _______________________________________________________________________________________________________________________________________
Verifizierer::~Verifizierer() {

}

// _______________________________________________________________________________________________________________________________________
void Verifizierer::startVerification(const string& sigFile) {
	
	double timeG = 0.0;
	double tstart = clock();
	
	// Initialize polynom from signature file.
	this->poly = Polynom(this->circuit.edgeIndex.size());  // Create polynom object with enough variable space.
	this->poly.addPolynom(this->createSignaturePolynom(this->readSignatureFile(sigFile)));	// Add polynom obtained from signature file.
	
	cout << "Anfangspolynom hat die Größe: " << this->poly.size() << endl;
	
	this->executeReplacements(this->makeReplaceList(this->createReplaceOrder()));
	
	if (this->poly.size() < 100) cout << "Endpolynom ist: " << this->poly <<  endl;
	cout << "mit Anzahl Momonem: " << this->poly.size() << endl;
	
	timeG += clock() - tstart; // Zeitmessung endet.
	timeG = timeG/CLOCKS_PER_SEC;
	cout << "Complete Verification process needed time: " << timeG << endl;

}

// _______________________________________________________________________________________________________________________________________
std::pair<std::map<std::string, std::vector<mpz_class>>, std::vector<string>> Verifizierer::readSignatureFile(const string& fileName) {
	// Verify file first.
	bool verify;
	ifstream in(fileName.c_str());
	if (!in) {
		verify = false;
	} else {
		verify = true;
	} 
	in.close();	
	if (!verify) throw vp::ParseError("File not found");
	
	// Create map to save signature coefficients and vector to save parts of the polynom description string.
	map<string, vector<mpz_class>> coff;
	vector<string> polyDescription;
	
	// Parse signature file line by line.
	string line;
	ifstream inputFile(fileName.c_str());
	inputFile >> line;
	while (line != "Signature:") inputFile >> line;
	getline(inputFile, line);	// Take next line, at the moment we are in the key word "Signature:" line.
	//while (line != "Polynom Description:") {
	// Help variables.
	size_t posEqual;
	string signalName;
	string coefficients;
	mpz_class currNum;
	size_t open;
	size_t close;
	vector<mpz_class> signalCoeff;
	// Filling the map with coefficients.
	while (getline(inputFile, line)) {
		if (line == "Polynom Description:") {  // Read in polynom description to build the polynom from it later on.
			getline(inputFile, line);  // Take next line with polynom description.
			while(line.find("+") != string::npos) {  // As long as where are plus signs, which are delimeters of polynom.
				size_t posPlus = line.find("+");
				string part = line.substr(0, posPlus);
				while (part.find(" ") != string::npos) {	// Delete spaces from part string.
					part.erase(part.find(" "), 1);
				}
				polyDescription.push_back(part);
				line.erase(0, posPlus+1);
				//cout << "Line:" << line << " , part:" << part << ";" << endl; 
			}
			while (line.find(" ") != string::npos) {	// Last summand.
				line.erase(line.find(" "), 1);
			}
			polyDescription.push_back(line);
			
			continue;
		}
		posEqual = line.find("="); 
		if (posEqual != string::npos) {
			// Get signal name.
			signalName = line.substr(0, posEqual);
			while (signalName.find(" ") != string::npos) {	// Delete spaces from signalName string.
				signalName.erase(signalName.find(" "), 1);
			} 
		} else {
			continue;
		}
		// Get coeffiecients.
		coefficients = line.substr(posEqual+1);
		open = coefficients.find("("); 
		close = coefficients.find(")");
		coefficients = coefficients.substr(open+1, close - open - 1);
		while (coefficients.find(" ") != string::npos) {	// Delete spaces from coefficients string.
			coefficients.erase(coefficients.find(" "), 1);
		} 
		while (coefficients.find(",") != string::npos) {  // Find number. Erase comma. Find next number, and so on...
			//currNum = std::stoi(coefficients.substr(0, coefficients.find(",")));  // Need mpz_class conversion.
			try {  // Check if coefficient is correct number. Otherwise terminate program. 
    				bool negative = false;	
    				if (coefficients.substr(0, coefficients.find(",")).find("-") != string::npos) {
    					negative = true;
    					coefficients.erase(0, 1);
    				}
    				int base = 2;
   				int expo = stoi(coefficients.substr(0, coefficients.find(",")));
    				currNum = pow(base, expo);
    				if (negative) currNum = currNum * -1; 
    				//currNum = mpz_class(coefficients.substr(0, coefficients.find(",")));
			}
			catch( const std::invalid_argument& e ) {
    				cerr <<  "std::invalid_argument exception caught at: " << e.what() << endl;
    				cout << "Wrong format in signature file. Stop program." << endl;
    				exit(EXIT_FAILURE);	
			}
			//currNum = mpz_class(coefficients.substr(0, coefficients.find(",")));
			coefficients.erase(0, coefficients.find(",") + 1);
			signalCoeff.push_back(currNum);
		}
		try {  // Check if last coefficient is correct number. Otherwise terminate program. 
   			bool negative = false;	
    			if (coefficients.find("-") != string::npos) {
    				negative = true;
    				coefficients.erase(0, 1);
    			}
    			int base = 2;
   			int expo(stoi(coefficients));
    			currNum = pow(base, expo);
    			if (negative) currNum = currNum * -1;
    			//currNum = mpz_class(coefficients);
		}
		catch( const std::invalid_argument& e ) {
    			cerr <<  "std::invalid_argument exception caught at: " << e.what() << endl;
    			cout << "Wrong format in signature file. Stop program." << endl;
    			exit(EXIT_FAILURE);	
		}
		//currNum = mpz_class(coefficients); // Add last number.
		signalCoeff.push_back(currNum);
		// Add signalCoeff in coff map with string signalName and coefficients in signalCoeff. Clear signalCoeff afterwards.
		coff.insert(pair<string, vector<mpz_class>>(signalName, signalCoeff));
		signalCoeff.clear();
	}
	
	// Show all input and output variables with given coefficients.
	//for (std::map<std::string, std::vector<mpz_class>>::iterator it = coff.begin(); it != coff.end(); ++it) {
	//	cout << "Coff: " << it->first << endl;
	//	for (auto& elem: it->second) cout << elem << endl;
	//}


	std::pair<std::map<std::string, std::vector<mpz_class>>, std::vector<string>> retValue;
	retValue.first = coff;
	retValue.second = polyDescription;
	return retValue;
}

// _______________________________________________________________________________________________________________________________________
Polynom Verifizierer::createSignaturePolynom(std::pair<std::map<std::string, std::vector<mpz_class>>, std::vector<string>> readValues) {
	Polynom p1(this->circuit.edgeIndex.size());  
	
	// Create polynom where signature polynom will be add and returned.
	
	std::map<std::string, std::vector<mpz_class>> coff = readValues.first;
	std::vector<string> description = readValues.second;
	
	// Make a list of all outputNode and inputNode names and at parallel a list of their edge indices.
	std::vector<string> primaryNames;
	std::vector<varIndex> primaryIndices;
	for (int i = 0; i < this->circuit.getOutputNodesCount(); i++) {  // Add all output node names and corresponding indices.
		 if (this->circuit.outputNode(i).type == "OUTPUT_PORT") {
		 	primaryNames.push_back(this->circuit.outputNode(i).name);
		 	primaryIndices.push_back(this->circuit.edgeIndex.at(*this->circuit.outputNode(i).inputs.at(0)));
		 } else {
		 	cout << "Something went wrong. This node is not an output node." << endl;
		 } 
	}
	for (int i = 0; i < this->circuit.getInputNodesCount(); i++) {  // Add all input node names and corresponding indices.
		 if (this->circuit.inputNode(i).type == "INPUT_PORT") {
		 	if (this->circuit.inputNode(i).name == "zeroNode") continue;
		 	if (this->circuit.inputNode(i).name == "oneNode") continue;
		 	//cout << "Input name: " << this->circuit.inputNode(i).name << endl;
		 	primaryNames.push_back(this->circuit.inputNode(i).name);
		 	//cout << "Input index: " << this->circuit.edgeIndex.at(*this->circuit.inputNode(i).outputs.at(0)) << endl;
		 	primaryIndices.push_back(this->circuit.edgeIndex.at(*this->circuit.inputNode(i).outputs.at(0)));
		 } else {
		 	cout << "Something went wrong. This node is not an input node." << endl;
		 }
	}
	
	if (primaryNames.size() != primaryIndices.size()) {  // Check whether both vectors have same size.
		cout << " Vector sizes do not match. Something went wrong. " << endl;
		return p1;
	}
	// Match all coefficients which were read in signature file and create the signature polynomial.
	map<string, Polynom> polynomials;
	Polynom currPolynom(this->circuit.edgeIndex.size());
	Polynom emptyPolynom(this->circuit.edgeIndex.size());
	string name = "";
	string lastName;
	for (int i = 0; i < primaryNames.size(); i++) {
		lastName = name;
		name = primaryNames.at(i).substr(0, primaryNames.at(i).find("[")); 
		if (coff.count(name) == 0) {  // Check primary signal name. If it is not given in signature file, continue with next.
			cout << "Primary signal with name: " << name << " not specified in signature." << endl;
			continue;
		}
		if (i > 0 && lastName != name) {
			polynomials.insert(pair<string, Polynom>(lastName, currPolynom));
			currPolynom = emptyPolynom;
		}
		string bitNumber = primaryNames.at(i).substr(primaryNames.at(i).find("[") + 1, primaryNames.at(i).find("]") - primaryNames.at(i).find("[") - 1);
		bool hasBitNumber = true;
		try {  // Check if signal has a bitNumber. 
    			std::stoi(bitNumber);
		}
		catch( const std::invalid_argument& e ) {
    			hasBitNumber = false;
    			
		}
		mpz_class factor = 0;
		if (hasBitNumber) {
			if (std::stoi(bitNumber) >= coff.at(name).size()) continue;  // This bit of signal was not specified in signature.
			factor = coff.at(name).at(std::stoi(bitNumber));
		} else {
			if (coff.at(name).size() == 0) continue;  // The one bit signal was not specified in signature. 
			factor = coff.at(name).at(0);
		}
		// Add monoms to currPolynom with factor given by signature file.
		if (factor == 0) continue;
		Monom2 temp(primaryIndices.at(i));
		temp.setFactor(factor);
		currPolynom.addMonom(temp);  // Add monoms with right indices and factor to temporary polynom.
	} 
	polynomials.insert(pair<string, Polynom>(name, currPolynom)); // Add last name and polynom part.
	// Polynom which represents simply the factor -1. Used for negation if a minus sign is detected in signature description.
	Monom2 minusOne;
	minusOne.setFactor(-1);
	Polynom negation(1);
	negation.addMonom(minusOne);
	Polynom one(1); 
	minusOne.setFactor(1);
	one.addMonom(minusOne);
	// Check the polynom description string and add single or multiplied signals to the signature polynom.
	for (vector<string>::iterator it = description.begin(); it != description.end(); ++it) {
		//cout << "String part: !" << *it << "!" << endl;
		if ((*it).find("*") != string::npos) {  // Found multiplication in polynom description.
			vector<string> parts; 
			if ((*it).find("-") != string::npos) {  // Negative version.
				string noMinus = (*it).erase((*it).find("-"), 1);
				while (noMinus.find("*") != string::npos) {  // Save every factor of multiplication.
					if (polynomials.count(noMinus.substr(0, noMinus.find("*"))) > 0) parts.push_back(noMinus.substr(0, noMinus.find("*")));
					noMinus.erase(0, noMinus.find("*")+1);
				}
				if (polynomials.count(noMinus) > 0) parts.push_back(noMinus);  // Add last factor.
				// Create multiplied polynom and add it to p1.
				Polynom tempPol;
				tempPol = Polynom::multiplyPoly(negation, polynomials.at(parts.back()));
				parts.pop_back();
				for (vector<string>::iterator partIte = parts.begin(); partIte != parts.end(); ++partIte) {
					tempPol = Polynom::multiplyPoly(tempPol, polynomials.at(*partIte));
				}
				p1.addPolynom(tempPol);
			} else {  // Positive version.
				string noMinus = (*it);
				while (noMinus.find("*") != string::npos) {  // Save every factor of multiplication.
					if (polynomials.count(noMinus.substr(0, noMinus.find("*"))) > 0) parts.push_back(noMinus.substr(0, noMinus.find("*")));
					noMinus.erase(0, noMinus.find("*")+1);
				}
				if (polynomials.count(noMinus) > 0) parts.push_back(noMinus);  // Add last factor.
				// Create multiplied polynom and add it to p1.
				Polynom tempPol;
				tempPol = Polynom::multiplyPoly(one, polynomials.at(parts.back()));
				parts.pop_back();
				for (vector<string>::iterator partIte = parts.begin(); partIte != parts.end(); ++partIte) {
					tempPol = Polynom::multiplyPoly(tempPol, polynomials.at(*partIte));
				}
				p1.addPolynom(tempPol);
			}
		} else {  // No multiplication in this part of polynom description. Just add partial polynom to p1.
			if ((*it).find("-") != string::npos) {  // Negative version.
				string noMinus = (*it).erase((*it).find("-"), 1);
				if (polynomials.count(noMinus) > 0) p1.addPolynom(Polynom::multiplyPoly(negation, polynomials.at(noMinus)));
			} else {  // Positive version.
				if (polynomials.count(*it) > 0) p1.addPolynom(polynomials.at(*it));
			}
		}
	}
	
	if (coff.count("constant_monom") > 0) {  // Signature file gives a constant monom factor.
		if (coff.at("constant_monom").at(0) != 0) { 
			Monom2 empty;
			empty.setFactor(coff.at("constant_monom").at(0));
			p1.addMonom(empty);
		}
	}
	return p1;
}

// _______________________________________________________________________________________________________________________________________
std::vector<vp::Node> Verifizierer::createReplaceOrder() {
	// Save replace order of nodes in this vector.
	std::vector<vp::Node> order;
	// Consider all nodes except for output and input nodes or deleted nodes.
	for (int i = 0; i < this->circuit.sortedNodes.size(); i++) {
		if (this->circuit.sortedNodes.at(i).type == "INPUT_PORT" || this->circuit.sortedNodes.at(i).type == "OUTPUT_PORT" || this->circuit.sortedNodes.at(i).type == "DELETED") {
		} else {
			order.push_back(this->circuit.sortedNodes.at(i));
		}
	}
	std::cout << "createReplaceOrder successfull." << std::endl;
	return order;
}

// _______________________________________________________________________________________________________________________________________
std::vector<replacement> Verifizierer::makeReplaceList(std::vector<vp::Node> order) {
	cout << "makeReplaceList start." << endl;
	std::vector<replacement> replaceList;  // Save all replacements which will be used later for backward rewriting.
	for (int i = 0; i < order.size(); i++) { 
		replaceList.push_back(convertNodeToReplacement(order.at(i)));
	}
	std::cout << "makeReplaceList successfull." << std::endl;
	return replaceList;
}

// _______________________________________________________________________________________________________________________________________
replacement Verifizierer::convertNodeToReplacement(vp::Node node) {
	replacement replace;
	replace.type = node.type;
	replace.output = this->circuit.edgeIndex.at(*node.outputs.at(0));
	replace.inputOne = this->circuit.edgeIndex.at(*node.inputs.at(0));
	if (node.type == std::string("not") || node.type == std::string("buffer") ) {
		replace.inputTwo = -1; // does not care since node has only one input.
	} else {
	replace.inputTwo = this->circuit.edgeIndex.at(*node.inputs.at(1));
	}
	return replace;
}

// _______________________________________________________________________________________________________________________________________
void Verifizierer::executeReplacements(std::vector<replacement> replacements) {
	int step = 1;
	double time1 = 0.0, tstart;
	time1 = 0.0;
	tstart = clock(); // Zeitmessung beginnt.
	// Execute all replacements given in replacement list.
	for (auto& elem: replacements) {
		//cout << "Step: " << step << ", Replacement Typ: " << elem.type << " with Vars: " << elem.output << ", " << elem.inputOne << ", " << elem.inputTwo << endl;
		if (this->poly.refList[elem.output].getSize() == 0) {
			//cout << "Variable not in polynom. No replacement needed." << endl;
			step++;
			continue;
		}	
		if (elem.type == "and") {
			//cout << "AND in " << elem.output << endl;
			if (elem.inputOne < elem.inputTwo) {
				this->poly.replaceAND(elem.output, elem.inputOne, elem.inputTwo);
			} else {
				this->poly.replaceAND(elem.output, elem.inputTwo, elem.inputOne);
			}
			
		} else if (elem.type == "xor")  {
			//cout << "XOR in " << elem.output << endl;
			if (elem.inputOne < elem.inputTwo) {
				this->poly.replaceXOR(elem.output, elem.inputOne, elem.inputTwo);
			} else {
				this->poly.replaceXOR(elem.output, elem.inputTwo, elem.inputOne);
			}
		} else if (elem.type == "or") {
			//cout << "OR in " << elem.output << endl;
			if (elem.inputOne < elem.inputTwo) {
				this->poly.replaceOR(elem.output, elem.inputOne, elem.inputTwo);
			} else {
				this->poly.replaceOR(elem.output, elem.inputTwo, elem.inputOne);
			}
		} else if (elem.type == "not") {
			//cout << "NOT in " << elem.output << endl;
			this->poly.replaceNOT(elem.output, elem.inputOne);
		} else if (elem.type == "buffer") {
			//cout << "NOT in " << elem.output << endl;
			this->poly.replaceBUFFER(elem.output, elem.inputOne);
		}
		cout << "Anzahl Monomen: " << this->poly.size() << endl;
		step++;
		//if (elem.type == "not" && elem.output < 130) cout << "Zwischenpolynom ist: " << this->poly << endl;
		//if (true) cout << "Zwischenpolynom ist: " << this->poly << endl;
	}
	time1 += clock() - tstart; // Zeitmessung endet.
	time1 = time1/CLOCKS_PER_SEC;
	cout << "Replacement operation needed time in sec. : " << time1 << endl;
}

