/*
 * Verifizierer.h
 *
 *  Created on: 08.01.2019
 *      Author: Alexander Konrad
 */

#include <string>
#include <stdlib.h>
#include <cstring>
#include <iostream>
#include <sstream>
#include <gmpxx.h>
#include <stdexcept>
#include <unordered_set>
#include "MyList.h"
#include "Circuit.h"
#include "Polynom.h"
#include "ParseError.h"
#include "math.h"
#include "time.h"


#ifndef VERIFIZIERER_H_
#define VERIFIZIERER_H_

typedef int varIndex;
struct replacement {
	string type;
	varIndex output;
	varIndex inputOne;
	varIndex inputTwo;
};

class Verifizierer {

//friend class Polynom;
public:	
	Verifizierer(const string& fileName, const string& sigFile);  // Construct verifier instance from a parsed circuit file and signature file.
	virtual ~Verifizierer();
	std::vector<vp::Node> createReplaceOrder();  // Create ordered vector of nodes, in this order which they will be replaced.
	std::vector<replacement> makeReplaceList(std::vector<vp::Node> order);
	replacement convertNodeToReplacement(vp::Node node);  // Convert a given node to a replacement struct.
	void executeReplacements(std::vector<replacement> replacements);  // Here all the replacements take place.
	void startVerification(const string& sigFile);  // Start the verification process without additional steps.

	std::pair<std::map<std::string, std::vector<mpz_class>>, std::vector<string>> readSignatureFile(const string& fileName);  // Read in the factors for the given inputs and outputs to create the signature polynom from in next step.
	Polynom createSignaturePolynom(std::pair<std::map<std::string, std::vector<mpz_class>>, std::vector<string>> readValue);  // Create the signature polynom given in the signature file.  
 
public:
	vp::Circuit circuit;
	Polynom poly;	

};

#endif /* VERIFIZIERER_H_ */
